@@ -38,34 +38,15 @@ dpdk_init(const struct smap *ovs_other_config)
     }
 }
 
+bool
+dpdk_attach_thread(unsigned cpu OVS_UNUSED)
+{
+    return false;
+}
+
 void
-dpdk_set_lcore_id(unsigned cpu OVS_UNUSED)
+dpdk_detach_thread(void)
 {
-    /* Nothing */
-}
-
-const char *
-dpdk_get_vhost_sock_dir(void)
-{
-    return NULL;
-}
-
-bool
-dpdk_vhost_iommu_enabled(void)
-{
-    return false;
-}
-
-bool
-dpdk_vhost_postcopy_enabled(void)
-{
-    return false;
-}
-
-bool
-dpdk_per_port_memory(void)
-{
-    return false;
 }
 
 bool
