@@ -35,10 +35,10 @@
 
 VLOG_DEFINE_THIS_MODULE(odp_execute_avx512);
 
-/* The below three build asserts make sure that l2_5_ofs, l3_ofs, and l4_ofs
- * fields remain in the same order and offset to l2_padd_size. This is needed
- * as the avx512_dp_packet_resize_l2() function will manipulate those fields at
- * a fixed memory index based on the l2_padd_size offset. */
+/* The below build asserts make sure that the below fields remain in the same
+ * order and offset to l2_pad_size. This is needed as the
+ * avx512_dp_packet_resize_l2() function will manipulate those fields at a
+ * fixed memory index based on the l2_pad_size offset. */
 BUILD_ASSERT_DECL(offsetof(struct dp_packet, l2_pad_size) +
                   MEMBER_SIZEOF(struct dp_packet, l2_pad_size) ==
                   offsetof(struct dp_packet, l2_5_ofs));
@@ -51,6 +51,14 @@ BUILD_ASSERT_DECL(offsetof(struct dp_packet, l3_ofs) +
                            MEMBER_SIZEOF(struct dp_packet, l3_ofs) ==
                            offsetof(struct dp_packet, l4_ofs));
 
+BUILD_ASSERT_DECL(offsetof(struct dp_packet, l4_ofs) +
+                           MEMBER_SIZEOF(struct dp_packet, l4_ofs) ==
+                           offsetof(struct dp_packet, inner_l3_ofs));
+
+BUILD_ASSERT_DECL(offsetof(struct dp_packet, inner_l3_ofs) +
+                           MEMBER_SIZEOF(struct dp_packet, inner_l3_ofs) ==
+                           offsetof(struct dp_packet, inner_l4_ofs));
+
 /* The below build assert makes sure it's safe to read/write 128-bits starting
  * at the l2_pad_size location. */
 BUILD_ASSERT_DECL(sizeof(struct dp_packet) -
@@ -112,7 +120,7 @@ avx512_dp_packet_resize_l2(struct dp_packet *b, int resize_by_bytes)
         dp_packet_pull(b, -resize_by_bytes);
     }
 
-    /* The next step is to update the l2_5_ofs, l3_ofs and l4_ofs fields which
+    /* The next step is to update the l2_5_ofs to inner_l4_ofs fields which
      * the scalar implementation does with the  dp_packet_adjust_layer_offset()
      * function. */
 
@@ -122,13 +130,14 @@ avx512_dp_packet_resize_l2(struct dp_packet *b, int resize_by_bytes)
     /* Set the v_u16_max register to all one's. */
     const __m128i v_u16_max = _mm_cmpeq_epi16(v_zeros, v_zeros);
 
-    /* Each lane represents 16 bits in a 12-bit register. In this case the
-     * first three 16-bit values, which will map to the l2_5_ofs, l3_ofs and
-     * l4_ofs fields. */
-    const uint8_t k_lanes = 0b1110;
+    /* Each lane represents 16 bits in a 128-bit register. Here the bitmask
+     * starts at l2_5_ofs with a value of 0 indicating it is not modified. Then
+     * five 1's to indicate modificaiton of all fields from l2_5_ofs to
+     * inner_l4_ofs. */
+    const uint8_t k_lanes = 0b111110;
 
     /* Set all 16-bit words in the 128-bits v_offset register to the value we
-     * need to add/substract from the l2_5_ofs, l3_ofs, and l4_ofs fields. */
+     * need to add/substract from the l2_5_ofs to inner_l4_ofs fields. */
     __m128i v_offset = _mm_set1_epi16(abs(resize_by_bytes));
 
     /* Load 128 bits from the dp_packet structure starting at the l2_pad_size
@@ -147,7 +156,7 @@ avx512_dp_packet_resize_l2(struct dp_packet *b, int resize_by_bytes)
     /* Based on the bytes adjust (positive, or negative) it will do the actual
      * add or subtraction. These functions will only operate on the lanes
      * (fields) requested based on k_cmp, i.e:
-     *   k_cmp = [l2_5_ofs, l3_ofs, l4_ofs]
+     *   k_cmp = [l2_5_ofs, ..., inner_l4_ofs]
      *   for field in kcmp
      *       v_adjust_src[field] = v_adjust_src[field] + v_offset
      */
@@ -450,7 +459,6 @@ action_avx512_ipv4_set_addrs(struct dp_packet_batch *batch,
 
     DP_PACKET_BATCH_FOR_EACH (i, packet, batch) {
         struct ip_header *nh = dp_packet_l3(packet);
-        ovs_be16 old_csum = ~nh->ip_csum;
 
         /* Load the 20 bytes of the IPv4 header. Without options, which is the
          * most common case it's 20 bytes, but can be up to 60 bytes. */
@@ -463,13 +471,20 @@ action_avx512_ipv4_set_addrs(struct dp_packet_batch *batch,
          * (v_pkt_masked). */
         __m256i v_new_hdr = _mm256_or_si256(v_key_shuf, v_pkt_masked);
 
-        /* Update the IP checksum based on updated IP values. */
-        uint16_t delta = avx512_ipv4_hdr_csum_delta(v_packet, v_new_hdr);
-        uint32_t new_csum = old_csum + delta;
-        delta = csum_finish(new_csum);
+        if (dp_packet_hwol_tx_ip_csum(packet)) {
+            dp_packet_ol_reset_ip_csum_good(packet);
+        } else {
+            ovs_be16 old_csum = ~nh->ip_csum;
 
-        /* Insert new checksum. */
-        v_new_hdr = _mm256_insert_epi16(v_new_hdr, delta, 5);
+            /* Update the IP checksum based on updated IP values. */
+            uint16_t delta = avx512_ipv4_hdr_csum_delta(v_packet, v_new_hdr);
+            uint32_t new_csum = old_csum + delta;
+
+            delta = csum_finish(new_csum);
+
+            /* Insert new checksum. */
+            v_new_hdr = _mm256_insert_epi16(v_new_hdr, delta, 5);
+        }
 
         /* If ip_src or ip_dst has been modified, L4 checksum needs to
          * be updated too. */
@@ -480,9 +495,11 @@ action_avx512_ipv4_set_addrs(struct dp_packet_batch *batch,
             size_t l4_size = dp_packet_l4_size(packet);
 
             if (nh->ip_proto == IPPROTO_UDP && l4_size >= UDP_HEADER_LEN) {
-                /* New UDP checksum. */
                 struct udp_header *uh = dp_packet_l4(packet);
-                if (uh->udp_csum) {
+                if (dp_packet_hwol_l4_is_udp(packet)) {
+                    dp_packet_ol_reset_l4_csum_good(packet);
+                } else if (uh->udp_csum) {
+                    /* New UDP checksum. */
                     uint16_t old_udp_checksum = ~uh->udp_csum;
                     uint32_t udp_checksum = old_udp_checksum + delta_checksum;
                     udp_checksum = csum_finish(udp_checksum);
@@ -495,13 +512,17 @@ action_avx512_ipv4_set_addrs(struct dp_packet_batch *batch,
                 }
             } else if (nh->ip_proto == IPPROTO_TCP &&
                        l4_size >= TCP_HEADER_LEN) {
-                /* New TCP checksum. */
-                struct tcp_header *th = dp_packet_l4(packet);
-                uint16_t old_tcp_checksum = ~th->tcp_csum;
-                uint32_t tcp_checksum = old_tcp_checksum + delta_checksum;
-                tcp_checksum = csum_finish(tcp_checksum);
+                if (dp_packet_hwol_l4_is_tcp(packet)) {
+                    dp_packet_ol_reset_l4_csum_good(packet);
+                } else {
+                    /* New TCP checksum. */
+                    struct tcp_header *th = dp_packet_l4(packet);
+                    uint16_t old_tcp_checksum = ~th->tcp_csum;
+                    uint32_t tcp_checksum = old_tcp_checksum + delta_checksum;
+                    tcp_checksum = csum_finish(tcp_checksum);
 
-                th->tcp_csum = tcp_checksum;
+                    th->tcp_csum = tcp_checksum;
+                }
             }
 
             pkt_metadata_init_conn(&packet->md);
@@ -563,11 +584,22 @@ avx512_ipv6_sum_header(__m512i ip6_header)
 
 static inline uint16_t ALWAYS_INLINE
 __attribute__((__target__("avx512vbmi")))
-avx512_ipv6_addr_csum_delta(__m512i old_header, __m512i new_header)
+avx512_ipv6_addr_csum_delta(__m512i v_packet, __m512i v_new_hdr,
+                            bool rh_present)
 {
-    uint16_t old_delta = avx512_ipv6_sum_header(old_header);
-    uint16_t new_delta = avx512_ipv6_sum_header(new_header);
-    uint32_t csum_delta = ((uint16_t) ~old_delta) + new_delta;
+    __m512i v_new_hdr_for_cksum = v_new_hdr;
+    uint32_t csum_delta;
+    uint16_t old_delta;
+    uint16_t new_delta;
+
+    if (rh_present) {
+        v_new_hdr_for_cksum = _mm512_mask_blend_epi64(0x18, v_new_hdr,
+                                                      v_packet);
+    }
+
+    old_delta = avx512_ipv6_sum_header(v_packet);
+    new_delta = avx512_ipv6_sum_header(v_new_hdr_for_cksum);
+    csum_delta = ((uint16_t) ~old_delta) + new_delta;
 
     return ~csum_finish(csum_delta);
 }
@@ -650,25 +682,19 @@ action_avx512_set_ipv6(struct dp_packet_batch *batch, const struct nlattr *a)
 
         if (do_csum) {
             size_t l4_size = dp_packet_l4_size(packet);
-            __m512i v_new_hdr_for_cksum = v_new_hdr;
             uint16_t delta_checksum;
 
-            /* In case of routing header being present, checksum should not be
-             * updated for the destination address. */
-            if (rh_present) {
-                v_new_hdr_for_cksum = _mm512_mask_blend_epi64(0x18, v_new_hdr,
-                                                              v_packet);
-            }
-
-            delta_checksum = avx512_ipv6_addr_csum_delta(v_packet,
-                                                         v_new_hdr_for_cksum);
-
             if (proto == IPPROTO_UDP && l4_size >= UDP_HEADER_LEN) {
                 struct udp_header *uh = dp_packet_l4(packet);
-
-                if (uh->udp_csum) {
+                if (dp_packet_hwol_l4_is_udp(packet)) {
+                    dp_packet_ol_reset_l4_csum_good(packet);
+                } else if (uh->udp_csum) {
+                    delta_checksum = avx512_ipv6_addr_csum_delta(v_packet,
+                                                                 v_new_hdr,
+                                                                 rh_present);
                     uint16_t old_udp_checksum = ~uh->udp_csum;
-                    uint32_t udp_checksum = old_udp_checksum + delta_checksum;
+                    uint32_t udp_checksum = old_udp_checksum +
+                                            delta_checksum;
 
                     udp_checksum = csum_finish(udp_checksum);
 
@@ -678,15 +704,26 @@ action_avx512_set_ipv6(struct dp_packet_batch *batch, const struct nlattr *a)
 
                     uh->udp_csum = udp_checksum;
                 }
-            } else if (proto == IPPROTO_TCP && l4_size >= TCP_HEADER_LEN) {
-                struct tcp_header *th = dp_packet_l4(packet);
-                uint16_t old_tcp_checksum = ~th->tcp_csum;
-                uint32_t tcp_checksum = old_tcp_checksum + delta_checksum;
 
-                tcp_checksum = csum_finish(tcp_checksum);
-                th->tcp_csum = tcp_checksum;
+            } else if (proto == IPPROTO_TCP && l4_size >= TCP_HEADER_LEN) {
+                if (dp_packet_hwol_l4_is_tcp(packet)) {
+                    dp_packet_ol_reset_l4_csum_good(packet);
+                } else {
+                    delta_checksum = avx512_ipv6_addr_csum_delta(v_packet,
+                                                                 v_new_hdr,
+                                                                 rh_present);
+                    struct tcp_header *th = dp_packet_l4(packet);
+                    uint16_t old_tcp_checksum = ~th->tcp_csum;
+                    uint32_t tcp_checksum = old_tcp_checksum + delta_checksum;
+
+                    tcp_checksum = csum_finish(tcp_checksum);
+                    th->tcp_csum = tcp_checksum;
+                }
             } else if (proto == IPPROTO_ICMPV6 &&
                        l4_size >= sizeof(struct icmp6_header)) {
+                delta_checksum = avx512_ipv6_addr_csum_delta(v_packet,
+                                                             v_new_hdr,
+                                                             rh_present);
                 struct icmp6_header *icmp = dp_packet_l4(packet);
                 uint16_t old_icmp6_checksum = ~icmp->icmp6_cksum;
                 uint32_t icmp6_checksum = old_icmp6_checksum + delta_checksum;
