@@ -65,8 +65,8 @@ static void
 oftrace_node_list_destroy(struct ovs_list *nodes)
 {
     if (nodes) {
-        struct oftrace_node *node, *next;
-        LIST_FOR_EACH_SAFE (node, next, node, nodes) {
+        struct oftrace_node *node;
+        LIST_FOR_EACH_SAFE (node, node, nodes) {
             ovs_list_remove(&node->node);
             oftrace_node_destroy(node);
         }
@@ -86,6 +86,7 @@ oftrace_node_destroy(struct oftrace_node *node)
 bool
 oftrace_add_recirc_node(struct ovs_list *recirc_queue,
                         enum oftrace_recirc_type type, const struct flow *flow,
+                        const struct ofpact_nat *ofn,
                         const struct dp_packet *packet, uint32_t recirc_id,
                         const uint16_t zone)
 {
@@ -101,6 +102,7 @@ oftrace_add_recirc_node(struct ovs_list *recirc_queue,
     node->flow = *flow;
     node->flow.recirc_id = recirc_id;
     node->flow.ct_zone = zone;
+    node->nat_act = ofn;
     node->packet = packet ? dp_packet_clone(packet) : NULL;
 
     return true;
@@ -179,6 +181,25 @@ oftrace_node_print_details(struct ds *output,
     }
 }
 
+static void
+oftrace_print_ip_flow(const struct flow *flow, int af, struct ds *output)
+{
+    if (af == AF_INET) {
+        ds_put_format(output, "nw_src="IP_FMT",tp_src=%"PRIu16","
+                      "nw_dst="IP_FMT",tp_dst=%"PRIu16,
+                      IP_ARGS(flow->nw_src), ntohs(flow->tp_src),
+                      IP_ARGS(flow->nw_dst), ntohs(flow->tp_dst));
+    } else if (af == AF_INET6) {
+        ds_put_cstr(output, "ipv6_src=");
+        ipv6_format_addr_bracket(&flow->ipv6_src, output, true);
+        ds_put_format(output, ",tp_src=%"PRIu16, ntohs(flow->tp_src));
+        ds_put_cstr(output, ",ipv6_dst=");
+        ipv6_format_addr_bracket(&flow->ipv6_dst, output, true);
+        ds_put_format(output, ",tp_dst=%"PRIu16, ntohs(flow->tp_dst));
+    }
+    ds_put_char(output, '\n');
+}
+
 /* Parses the 'argc' elements of 'argv', ignoring argv[0].  The following
  * forms are supported:
  *
@@ -192,7 +213,7 @@ parse_flow_and_packet(int argc, const char *argv[],
                       struct ofproto_dpif **ofprotop, struct flow *flow,
                       struct dp_packet **packetp,
                       struct ovs_list *next_ct_states,
-                      bool *consistent)
+                      bool *consistent, bool *names)
 {
     const struct dpif_backer *backer = NULL;
     char *error = NULL;
@@ -212,6 +233,9 @@ parse_flow_and_packet(int argc, const char *argv[],
     if (consistent) {
         *consistent = false;
     }
+    if (names) {
+        *names = false;
+    }
     for (int i = 1; i < argc; i++) {
         const char *arg = argv[i];
         if (!strcmp(arg, "-generate") || !strcmp(arg, "--generate")) {
@@ -249,6 +273,10 @@ parse_flow_and_packet(int argc, const char *argv[],
                    && (!strcmp(arg, "-consistent") ||
                        !strcmp(arg, "--consistent"))) {
             *consistent = true;
+        } else if (names
+                   && (!strcmp(arg, "-names") ||
+                       !strcmp(arg, "--names"))) {
+            *names = true;
         } else if (!strcmp(arg, "--ct-next")) {
             if (i + 1 >= argc) {
                 error = xasprintf("Missing argument for option %s", arg);
@@ -397,11 +425,7 @@ parse_flow_and_packet(int argc, const char *argv[],
         }
 
         struct ofputil_port_map map = OFPUTIL_PORT_MAP_INITIALIZER(&map);
-        const struct ofport *ofport;
-        HMAP_FOR_EACH (ofport, hmap_node, &(*ofprotop)->up.ports) {
-            ofputil_port_map_put(&map, ofport->ofp_port,
-                                 netdev_get_name(ofport->netdev));
-        }
+        ofproto_append_ports_to_map(&map, (*ofprotop)->up.ports);
         char *err = parse_ofp_exact_flow(flow, NULL,
                                          ofproto_get_tun_tab(&(*ofprotop)->up),
                                          args[n_args - 1], &map);
@@ -416,7 +440,7 @@ parse_flow_and_packet(int argc, const char *argv[],
     if (generate_packet) {
         /* Generate a packet, as requested. */
         packet = dp_packet_new(0);
-        flow_compose(packet, flow, l7, l7_len);
+        flow_compose(packet, flow, l7, l7_len, false);
     } else if (packet) {
         /* Use the metadata from the flow and the packet argument to
          * reconstruct the flow. */
@@ -453,16 +477,17 @@ ofproto_unixctl_trace(struct unixctl_conn *conn, int argc, const char *argv[],
     struct dp_packet *packet;
     char *error;
     struct flow flow;
+    bool names;
     struct ovs_list next_ct_states = OVS_LIST_INITIALIZER(&next_ct_states);
 
     error = parse_flow_and_packet(argc, argv, &ofproto, &flow, &packet,
-                                  &next_ct_states, NULL);
+                                  &next_ct_states, NULL, &names);
     if (!error) {
         struct ds result;
 
         ds_init(&result);
         ofproto_trace(ofproto, &flow, packet, NULL, 0, &next_ct_states,
-                      &result);
+                      &result, names);
         unixctl_command_reply(conn, ds_cstr(&result));
         ds_destroy(&result);
         dp_packet_delete(packet);
@@ -480,6 +505,7 @@ ofproto_unixctl_trace_actions(struct unixctl_conn *conn, int argc,
     enum ofputil_protocol usable_protocols;
     struct ofproto_dpif *ofproto;
     bool enforce_consistency;
+    bool names;
     struct ofpbuf ofpacts;
     struct dp_packet *packet;
     struct ds result;
@@ -509,7 +535,8 @@ ofproto_unixctl_trace_actions(struct unixctl_conn *conn, int argc,
     }
 
     error = parse_flow_and_packet(argc, argv, &ofproto, &match.flow, &packet,
-                                  &next_ct_states, &enforce_consistency);
+                                  &next_ct_states, &enforce_consistency,
+                                  &names);
     if (error) {
         unixctl_command_reply_error(conn, error);
         free(error);
@@ -557,7 +584,8 @@ ofproto_unixctl_trace_actions(struct unixctl_conn *conn, int argc,
     }
 
     ofproto_trace(ofproto, &match.flow, packet,
-                  ofpacts.data, ofpacts.size, &next_ct_states, &result);
+                  ofpacts.data, ofpacts.size, &next_ct_states, &result,
+                  names);
     unixctl_command_reply(conn, ds_cstr(&result));
 
 exit:
@@ -637,11 +665,78 @@ execute_actions_except_outputs(struct dpif *dpif,
     ofpbuf_uninit(&pruned_actions);
 }
 
+static void
+ofproto_trace_recirc_node(struct oftrace_recirc_node *node,
+                          struct ovs_list *next_ct_states,
+                          struct ds *output)
+{
+    ds_put_cstr(output, "\n\n");
+    ds_put_char_multiple(output, '=', 79);
+    ds_put_format(output, "\nrecirc(%#"PRIx32")", node->recirc_id);
+
+    if (next_ct_states && node->type == OFT_RECIRC_CONNTRACK) {
+        uint32_t ct_state;
+        if (ovs_list_is_empty(next_ct_states)) {
+            ct_state = CS_TRACKED | CS_NEW;
+            ds_put_cstr(output, " - resume conntrack with default "
+                        "ct_state=trk|new (use --ct-next to customize)");
+        } else {
+            ct_state = oftrace_pop_ct_state(next_ct_states);
+            struct ds s = DS_EMPTY_INITIALIZER;
+            format_flags(&s, ct_state_to_string, ct_state, '|');
+            ds_put_format(output, " - resume conntrack with ct_state=%s",
+                          ds_cstr(&s));
+            ds_destroy(&s);
+        }
+        node->flow.ct_state = ct_state;
+    }
+    ds_put_char(output, '\n');
+
+    /* If there's any snat/dnat information assume we always translate to
+     * the first IP/port to make sure we don't match on incorrect flows later
+     * on.
+     */
+    if (node->nat_act) {
+        const struct ofpact_nat *ofn = node->nat_act;
+
+        ds_put_cstr(output, "Replacing src/dst IP/ports to simulate NAT:\n");
+        ds_put_cstr(output, " Initial flow: ");
+        oftrace_print_ip_flow(&node->flow, ofn->range_af, output);
+
+        if (ofn->flags & NX_NAT_F_SRC) {
+            if (ofn->range_af == AF_INET) {
+                node->flow.nw_src = ofn->range.addr.ipv4.min;
+            } else if (ofn->range_af == AF_INET6) {
+                node->flow.ipv6_src = ofn->range.addr.ipv6.min;
+            }
+
+            if (ofn->range_af != AF_UNSPEC && ofn->range.proto.min) {
+                node->flow.tp_src = htons(ofn->range.proto.min);
+            }
+        }
+        if (ofn->flags & NX_NAT_F_DST) {
+            if (ofn->range_af == AF_INET) {
+                node->flow.nw_dst = ofn->range.addr.ipv4.min;
+            } else if (ofn->range_af == AF_INET6) {
+                node->flow.ipv6_dst = ofn->range.addr.ipv6.min;
+            }
+
+            if (ofn->range_af != AF_UNSPEC && ofn->range.proto.min) {
+                node->flow.tp_dst = htons(ofn->range.proto.min);
+            }
+        }
+        ds_put_cstr(output, " Modified flow: ");
+        oftrace_print_ip_flow(&node->flow, ofn->range_af, output);
+    }
+    ds_put_char_multiple(output, '=', 79);
+    ds_put_cstr(output, "\n\n");
+}
+
 static void
 ofproto_trace__(struct ofproto_dpif *ofproto, const struct flow *flow,
                 const struct dp_packet *packet, struct ovs_list *recirc_queue,
                 const struct ofpact ofpacts[], size_t ofpacts_len,
-                struct ds *output)
+                struct ds *output, bool names)
 {
     struct ofpbuf odp_actions;
     ofpbuf_init(&odp_actions, 0);
@@ -649,6 +744,8 @@ ofproto_trace__(struct ofproto_dpif *ofproto, const struct flow *flow,
     struct xlate_in xin;
     struct flow_wildcards wc;
     struct ovs_list trace = OVS_LIST_INITIALIZER(&trace);
+    struct ofputil_port_map map = OFPUTIL_PORT_MAP_INITIALIZER(&map);
+    struct hmap *portno_names = NULL;
     xlate_in_init(&xin, ofproto,
                   ofproto_dpif_get_tables_version(ofproto), flow,
                   flow->in_port.ofp_port, NULL, ntohs(flow->tcp_flags),
@@ -657,12 +754,27 @@ ofproto_trace__(struct ofproto_dpif *ofproto, const struct flow *flow,
     xin.ofpacts_len = ofpacts_len;
     xin.trace = &trace;
     xin.recirc_queue = recirc_queue;
+    xin.names = names;
+
+    if (names) {
+        ofproto_append_ports_to_map(&map, ofproto->up.ports);
+
+        portno_names = xmalloc(sizeof *portno_names);
+        hmap_init(portno_names);
+
+        struct dpif_port dpif_port;
+        struct dpif_port_dump port_dump;
+        DPIF_PORT_FOR_EACH (&dpif_port, &port_dump, ofproto->backer->dpif) {
+            odp_portno_names_set(portno_names, dpif_port.port_no,
+                                 dpif_port.name);
+        }
+    }
 
     /* Copy initial flow out of xin.flow.  It differs from '*flow' because
      * xlate_in_init() initializes actset_output to OFPP_UNSET. */
     struct flow initial_flow = xin.flow;
     ds_put_cstr(output, "Flow: ");
-    flow_format(output, &initial_flow, NULL);
+    flow_format(output, &initial_flow, &map);
     ds_put_char(output, '\n');
 
     struct xlate_out xout;
@@ -674,18 +786,19 @@ ofproto_trace__(struct ofproto_dpif *ofproto, const struct flow *flow,
     if (flow_equal(&initial_flow, &xin.flow)) {
         ds_put_cstr(output, "unchanged");
     } else {
-        flow_format(output, &xin.flow, NULL);
+        flow_format(output, &xin.flow, &map);
     }
     ds_put_char(output, '\n');
 
     ds_put_cstr(output, "Megaflow: ");
     struct match match;
     match_init(&match, flow, &wc);
-    match_format(&match, NULL, output, OFP_DEFAULT_PRIORITY);
+    match_format(&match, &map, output, OFP_DEFAULT_PRIORITY);
     ds_put_char(output, '\n');
 
     ds_put_cstr(output, "Datapath actions: ");
-    format_odp_actions(output, odp_actions.data, odp_actions.size, NULL);
+    format_odp_actions(output, odp_actions.data, odp_actions.size,
+                       portno_names);
 
     if (error != XLATE_OK) {
         ds_put_format(output,
@@ -702,6 +815,13 @@ ofproto_trace__(struct ofproto_dpif *ofproto, const struct flow *flow,
         }
     }
 
+    if (names) {
+        ofputil_port_map_destroy(&map);
+
+        odp_portno_names_destroy(portno_names);
+        hmap_destroy(portno_names);
+        free(portno_names);
+    }
 
     xlate_out_uninit(&xout);
     ofpbuf_uninit(&odp_actions);
@@ -721,41 +841,19 @@ void
 ofproto_trace(struct ofproto_dpif *ofproto, const struct flow *flow,
               const struct dp_packet *packet,
               const struct ofpact ofpacts[], size_t ofpacts_len,
-              struct ovs_list *next_ct_states, struct ds *output)
+              struct ovs_list *next_ct_states, struct ds *output,
+              bool names)
 {
     struct ovs_list recirc_queue = OVS_LIST_INITIALIZER(&recirc_queue);
     ofproto_trace__(ofproto, flow, packet, &recirc_queue,
-                    ofpacts, ofpacts_len, output);
+                    ofpacts, ofpacts_len, output, names);
 
     struct oftrace_recirc_node *recirc_node;
     LIST_FOR_EACH_POP (recirc_node, node, &recirc_queue) {
-        ds_put_cstr(output, "\n\n");
-        ds_put_char_multiple(output, '=', 79);
-        ds_put_format(output, "\nrecirc(%#"PRIx32")",
-                      recirc_node->recirc_id);
-
-        if (next_ct_states && recirc_node->type == OFT_RECIRC_CONNTRACK) {
-            uint32_t ct_state;
-            if (ovs_list_is_empty(next_ct_states)) {
-                ct_state = CS_TRACKED | CS_NEW;
-                ds_put_cstr(output, " - resume conntrack with default "
-                            "ct_state=trk|new (use --ct-next to customize)");
-            } else {
-                ct_state = oftrace_pop_ct_state(next_ct_states);
-                struct ds s = DS_EMPTY_INITIALIZER;
-                format_flags(&s, ct_state_to_string, ct_state, '|');
-                ds_put_format(output, " - resume conntrack with ct_state=%s",
-                              ds_cstr(&s));
-                ds_destroy(&s);
-            }
-            recirc_node->flow.ct_state = ct_state;
-        }
-        ds_put_char(output, '\n');
-        ds_put_char_multiple(output, '=', 79);
-        ds_put_cstr(output, "\n\n");
-
+        ofproto_trace_recirc_node(recirc_node, next_ct_states, output);
         ofproto_trace__(ofproto, &recirc_node->flow, recirc_node->packet,
-                        &recirc_queue, ofpacts, ofpacts_len, output);
+                        &recirc_queue, ofpacts, ofpacts_len, output,
+                        names);
         oftrace_recirc_node_destroy(recirc_node);
     }
 }
@@ -779,3 +877,13 @@ ofproto_dpif_trace_init(void)
         "[-generate|packet] actions",
         2, INT_MAX, ofproto_unixctl_trace_actions, NULL);
 }
+
+void
+ofproto_append_ports_to_map(struct ofputil_port_map *map, struct hmap ports) {
+    struct ofport *ofport;
+
+    HMAP_FOR_EACH (ofport, hmap_node, &ports) {
+        ofputil_port_map_put(map, ofport->ofp_port,
+                             netdev_get_name(ofport->netdev));
+    }
+}
